#include "ipc.h"
#include "page_mapper.h"



PageMapper::PageMapper(Accessor *accessor, Pages *pages)
: _accessor(accessor), _pages(pages), _attached(0)
{
}

/* Accounting methods. */

/* Attach a pager, opening the accessor if required. */

void PageMapper::attach()
{
    std::lock_guard<std::mutex> guard(_lock);

    if (!_attached)
        _accessor->open();

    _attached += 1;
}

/* Detach a pager, purging active pages and closing the accessor if no more
   pagers are attached. Return whether any pagers are still attached. */

unsigned int PageMapper::detach()
{
    std::lock_guard<std::mutex> guard(_lock);

    if (_attached)
    {
        _attached -= 1;

        if (!_attached)
        {
            _map.purge(this, _pages);
            _accessor->close();
        }
    }

    return _attached;
}

/* Interface for the pager. */

/* Return a flexpage providing access to the indicated file 'offset'.

   The returned flexpage will either be an existing, compatible flexpage or a
   completely new flexpage.

   This method locks the accessor to prevent concurrent queries with the
   same details, with the lock held until the queue or flush_all operations
   release the lock. */

Flexpage *PageMapper::get(offset_t offset)
{
    _lock.lock();

    Flexpage *f = find(offset);

    if (f == NULL)
        f = flexpage(offset);

    f->increment();
    return f;
}

/* Queue the given 'flexpage' in the page collection, making it available for
   eventual reuse.

   This method unlocks the accessor. */

void PageMapper::queue(Flexpage *flexpage)
{
    _pages->queue(this, flexpage);

    _lock.unlock();
}

/* Flush pages in the given range from 'start' with 'size'. */

void PageMapper::flush_all(offset_t start, offset_t size)
{
    std::lock_guard<std::mutex> guard(_lock);

    _map.flush_all(start, size, this, _pages);
}

/* Return the maximum extent of the mapped resource. */

offset_t PageMapper::get_data_size()
{
    return _accessor->get_size();
}

/* Internal flexpage retrieval methods. */

/* Find an existing flexpage for 'offset'. Where the accessor has registered a
   compatible flexpage, an attempt is made to reserve it in the page collection;
   if this succeeds, the flexpage is returned. Otherwise, NULL is returned. */

Flexpage *PageMapper::find(offset_t offset)
{
    Flexpage *flexpage = _map.find(offset);

    /* Between finding and reserving a flexpage, there is a possibility that
       another accessor might acquire the flexpage, issue it, and even purge
       it. */

    if ((flexpage != NULL) && _pages->reserve(this, flexpage))
        return flexpage;
    else
        return NULL;
}

/* Obtain a new flexpage for the file 'offset'. If the page collection is unable
   to obtain a completely new flexpage, an existing flexpage is requested from
   the page collection and recycled.

   The obtained flexpage is filled with content. */

Flexpage *PageMapper::flexpage(offset_t offset)
{
    Flexpage *flexpage = _pages->flexpage();

    /* Obtain an existing flexpage and reuse it. */

    if (flexpage == NULL)
        flexpage = _pages->remove();

    flexpage->reset(offset);

    fill(flexpage);
    _map.insert(flexpage);
    return flexpage;
}

/* Interface for the page collection. */

/* Remove the record of 'flexpage' in this accessor, flushing its content. */

void PageMapper::remove(Flexpage *flexpage)
{
    _map.remove(this, flexpage);
}

/* Data transfer methods. */

void PageMapper::fill(Flexpage *flexpage)
{
    _accessor->fill(flexpage);
}

void PageMapper::flush(Flexpage *flexpage, bool purge)
{
    if (flexpage->decrement() || purge)
    {
        _accessor->flush(flexpage);
        ipc_unmap_flexpage(flexpage);
        flexpage->invalidate();
    }
}

// vim: tabstop=4 expandtab shiftwidth=4
